<?php

declare(strict_types=1);

namespace apexl\Io\modules\user\interfaces;

use apexl\Io\modules\user\services\Permissions;

interface registersPermissions
{
    public function permissions(Permissions $permissions): void;
}
