<?php

declare(strict_types=1);

namespace apexl\Io\modules\user\traits;

use apexl\Utils\Class\ClassBasename;

trait IsPermissionTrait
{
    public function jsonSerialize(): string
    {
        return sprintf('%s__%s', ClassBasename::classBasename($this::class), $this->name);
    }
}
